import io
import logging
import os
from math import sqrt, floor
from PIL import Image
import numpy as np
import reedsolo

rsc = reedsolo.RSCodec(1)
CHANNELS = [1, 2, 4]

def encode_b128(b):
	paddings = 0
	while b[-1] == 0:
		paddings += 1
		b = b[:-1]
	import bitarray
	b1 = bitarray.bitarray(endian="big")
	b1.frombytes(b)
	b2 = bitarray.bitarray(endian="big")
	while len(b1):
		temp, b1 = b1[:7], b1[7:]
		temp.insert(0, 0)
		b2.extend(temp)
	return b2.tobytes() + b"\x00" * paddings

def decode_b128(b):
	assert b.isascii()
	import bitarray
	b1 = bitarray.bitarray(endian="big")
	b1.frombytes(b)
	b2 = bitarray.bitarray(endian="big")
	while len(b1):
		temp, b1 = b1[1:8], b1[8:]
		if len(temp) < 7:
			break
		b2.extend(temp)
	return b2.tobytes().removesuffix(b"\x00")

# text = b"Hello World!\n" * 100
# assert decode_b128(encode_b128(text)) == text
# assert encode_b128(decode_b128(text)).removesuffix(b"\x00") == text

def quantise_into(a, clip=None, in_place=True, checker=1, dtype=np.uint8):
	if issubclass(a.dtype.type, np.integer):
		return a
	if checker:
		inds = np.indices(a.shape, dtype=np.uint8).sum(axis=0, dtype=np.uint8)
		inds &= checker
		inds = inds.view(bool)
	z = np.random.random_sample(a.shape)
	if checker:
		z %= 0.5
	else:
		z %= 1 - 2 ** -12
	a = np.add(a, z, out=a if in_place else None)
	if checker:
		a[inds] += 0.5 - 2 ** -12
	if clip:
		a = np.clip(a, *clip, out=a)
	if issubclass(getattr(dtype, "type", dtype), np.floating):
		np.floor(a, out=a)
	return np.asanyarray(a, dtype=dtype)

def encode_nrzi(text):
	rs2 = reedsolo.RSCodec(min(len(text) + 1 >> 1, 85))
	data = rs2.encode(text)
	arr = np.frombuffer(data, dtype=np.uint8)
	bits = np.empty(len(data) * 8, dtype=np.uint8)
	for i in range(8):
		bits[i::8] = arr & (1 << i) != 0
	cum = np.cumsum(bits, out=bits)
	cum &= 1
	nrzi = cum.view(bool)
	return nrzi

def decode_nrzi(nrzi):
	diff = np.diff(nrzi.view(np.int8))
	bits = np.concatenate(([nrzi[0]], diff.view(bool))).astype(np.uint8)
	arr = np.zeros(len(bits) // 8, dtype=np.uint8)
	for i in range(8):
		arr += bits[i::8] << i
	data = arr.data
	rs2 = reedsolo.RSCodec(min(round(len(data) / 3), 85))
	text = rs2.decode(data)
	return text[0]

def encode_qr(text, version=3):
	import base45
	import pyqrcode
	text = base45.b45encode(text).upper().decode("ascii")
	size = len(text)
	if version == 3:
		assert size <= 77, size
		err = "L"
		w = 32
	elif version == 11:
		assert size <= 259, size
		err = "Q"
		w = 64
	elif version == 27:
		assert size <= 910, size
		err = "H"
		w = 128
	else:
		raise NotImplementedError(version)
	img = pyqrcode.create(text, error=err, version=version, mode="alphanumeric", encoding="ascii")
	import io
	b = io.BytesIO()
	img.png(b, scale=1, module_color=(0,) * 3, background=(255,) * 4, quiet_zone=2)
	b.seek(0)
	box = (0, 0, w, w)
	imo = Image.open(b)
	import random
	if random.randint(0, 1):
		imo = imo.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
	imo = imo.rotate(random.randint(0, 3) * 90)
	return imo.crop(box)

def encode_barcode(text):
	import barcode
	num = int.from_bytes(text, "big")
	writer = barcode.Code128(str(num), writer=barcode.writer.ImageWriter("png"))
	import io
	b = io.BytesIO()
	writer.write(b, {"font_size": 3, "text_distance": 3, "module_width": 1, "module_height": 24, "quiet_zone": 8, "dpi": 25.40000001})
	b.seek(0)
	im = Image.open(b)
	return im.crop((0, 1, im.width, 25))

reader1 = None
reader2 = None
def decode_qr(images, channels=1):
	global reader1, reader2
	if isinstance(images, Image.Image):
		images = [images]
	if not reader1:
		import pyzbar.pyzbar
		reader1 = pyzbar.pyzbar
	import base45
	seen = []
	for image in images:
		if image.mode not in ("L", "RGB"):
			image = image.convert("L")
		for scale_factor in (0.25, 1 / 3, 0.5):
			im = image.resize((round(image.width * scale_factor), round(image.height * scale_factor)), resample=Image.Resampling.LANCZOS)
			a = (np.asanyarray(im, dtype=np.uint8) > 127).view(np.uint8)
			a *= 255
			im = Image.fromarray(a, "L").resize((im.width * 3, im.height * 3), resample=Image.Resampling.NEAREST)
			dets = reader1.decode(im, symbols=[reader1.ZBarSymbol.QRCODE])
			if dets:
				a2 = np.array(image, dtype=np.uint8)
				polys = []
				for data in dets:
					decoded = data.data
					try:
						value = base45.b45decode(decoded)
					except ValueError:
						continue
					poly = np.array([[point.x / scale_factor / 3, point.y / scale_factor / 3] for point in data.polygon], dtype=np.int32)
					polys.append(poly)
					while len(value) >= 51:
						yield value[:51]
						value = value[51:]
				import cv2
				cv2.fillPoly(a2, pts=polys, color=(255, 255, 255))
				image = Image.fromarray(a2)
		if max(image.width, image.height) > 1024:
			small = min(image.width, image.height, 1024)
			image = image.resize((small, small), resample=Image.Resampling.LANCZOS)
		seen.append(image)
	if not reader2:
		from qreader import QReader
		reader2 = QReader(model_size="s", min_confidence=0.25, reencode_to=None)
	for im in tuple(seen):
		if im.width == im.height == 1024:
			quads = (
				im.crop((0, 0, 512, 512)).rotate(180),
				im.crop((512, 512, 1024, 1024)),
				im.crop((0, 512, 512, 1024)).rotate(90),
				im.crop((512, 0, 1024, 512)).rotate(-90),
			)
			seen.extend(quads)
	found = set()
	retries = []
	for i, image in enumerate(seen):
		image_np = np.asanyarray(image, dtype=np.uint8)
		for data in reader2.detect(image_np):
			if i not in found and image.width == image.height == 1024:
				quads = (
					image.crop((0, 0, 512, 512)).rotate(180),
					image.crop((512, 512, 1024, 1024)),
					image.crop((0, 512, 512, 1024)).rotate(90),
					image.crop((512, 0, 1024, 512)).rotate(-90),
				)
				retries.extend(quads)
			found.add(i)
			if not data:
				yield None
				continue
			try:
				decoded = reader2.decode(image_np, data)
			except Exception:
				import traceback
				traceback.print_exc()
				yield None
				continue
			if not decoded:
				yield None
				continue
			try:
				value = base45.b45decode(decoded)
			except ValueError:
				continue
			while len(value) >= 51:
				yield value[:51]
				value = value[51:]
	for i, image in enumerate(retries):
		image_np = np.asanyarray(image, dtype=np.uint8)
		for data in reader2.detect(image_np):
			if not data:
				yield None
				continue
			try:
				decoded = reader2.decode(image_np, data)
			except Exception:
				import traceback
				traceback.print_exc()
				yield None
				continue
			if not decoded:
				yield None
				continue
			try:
				value = base45.b45decode(decoded)
			except ValueError:
				continue
			while len(value) >= 51:
				yield value[:51]
				value = value[51:]

def encode_message(message):
	if isinstance(message, str):
		message = message.encode("utf-8")
	if isinstance(message, memoryview):
		message = bytes(message)
	elif not isinstance(message, bytes):
		raise TypeError("`message` should be an instance of `str`, `bytes`, or `memoryview`.")
	# import zlib
	# import lzma
	import brotli
	import paq
	encodes = {}
	encodes["raw"] = b"\x91" + message
	if message.isascii():
		encodes["b128"] = b"\x92" + decode_b128(message)
	# encodes["zlib"] = zlib.compress(message, level=9)
	# encodes["lzma"] = lzma.compress(message, format=lzma.FORMAT_ALONE, check=-1, preset=9, filters=None)
	encodes["brotli"] = brotli.compress(message, quality=11, mode=brotli.MODE_TEXT)
	encodes["paq"] = paq.compress(message)
	order = sorted(encodes, key=lambda k: len(encodes[k]))
	# print("Compressed sizes:\n" + "\n".join(f"{k}:{len(encodes[k])}" for k in order))
	return encodes[order[0]]

def decode_message(message):
	if isinstance(message, str):
		message = message.encode("utf-8")
	if isinstance(message, memoryview):
		message = bytes(message)
	elif not isinstance(message, bytes):
		raise TypeError("`message` should be an instance of `str`, `bytes`, or `memoryview`.")
	if not message:
		raise ValueError("Input is empty.")
	if message.startswith(b"\x00c"):
		import paq
		return paq.decompress(message)
	if message[:1] in b"\x0b\x1b\x8b":
		import brotli
		return brotli.decompress(message)
	# if message[:1] == b"]":
	# 	import lzma
	# 	return lzma.decompress(message)
	# if message[:1] == b"x":
	# 	import zlib
	# 	return zlib.decompress(message)
	if message[:1] == b"\x92":
		return encode_b128(message[1:]).removesuffix(b"\x00")
	if message[:1] == b"\x91":
		return message[1:]
	raise ValueError("Unrecognised encoding " + str(message[:12]))
# assert decode_message(encode_message(m := b"Hello World!\n" * 100)) == m

def diagonals(rows, cols):
	indices = np.empty((2, rows * cols), dtype=np.uint32)
	i = 0
	for d in range(rows + cols - 1):
		row_start = max(0, d - cols + 1)
		row_end = min(rows - 1, d)
		col_indices = np.arange(row_start, row_end + 1)
		row_indices = d - col_indices
		size = len(row_indices)
		indices[1][i:i + size] = row_indices
		indices[0][i:i + size] = col_indices
		i += size
	return indices.T
# assert np.all(diagonals(4, 3) == [[0, 0], [0, 1], [1, 0], [0, 2], [1, 1], [2, 0], [1, 2], [2, 1], [3, 0], [2, 2], [3, 1], [3, 2]])

def tesselate(tesselation, grid, tile, seed=0):
	W, H = grid
	w, h = tile
	rows = H // h
	cols = W // w
	if tesselation == "plain":
		coords = diagonals(rows, cols)
		coords *= np.uint32((w, h))
		for x, y in coords:
			yield (x, y, w, h)
		return
	if tesselation == "pythagorean":
		cells = np.ones((rows, cols), dtype=bool)
		init = -(seed & 3)
		row = 0
		col = init
		offs = 0
		while True:
			y = -col + (offs * 2) + offs // 2
			x = row * 2 + (offs & 1)
			if x + 1 >= rows or y < 0:
				offs += 1
				if offs * 2 >= cols * 1.5:
					break
				row = 0
				col = init
				continue
			if y < rows - 1:
				cells[y:y + 2].T[x:x + 2] = 0
				yield (x * w, y * h, w * 2, h * 2)
			row += 1
			col += 1
		small = np.nonzero(cells)
		for y, x in zip(*small):
			yield (x * w, y * h, w, h)
		return
	raise NotImplementedError(tesselation)

def batch_images(it, maxsize, tesselation="plain", tile=4, qr_scale=2, seed=0):
	def retrieve(qr_scale=2):
		try:
			qr_info = qr_map[qr_scale]
		except KeyError:
			raise NotImplementedError(qr_scale)
		if qr_scale == 2:
			text = next(it)
		elif qr_scale == 4:
			text = next(it) + next(it) + next(it)
		else:
			raise NotImplementedError(qr_scale)
		qr_version = qr_info["version"]
		return encode_qr(text, version=qr_version)
	im = retrieve(qr_scale=qr_scale)
	scaled = im.width * tile
	grid_img = Image.new("1", (maxsize, maxsize))
	rtes = tesselation if tesselation != "diagonal" else "plain"
	for x, y, w, h in tesselate(rtes, (maxsize, maxsize), (scaled, scaled), seed=seed):
		qs = 2
		# print(qs, w, h, scaled, scaled)
		im2 = retrieve(qs).resize((w, h), resample=Image.Resampling.NEAREST)
		grid_img.paste(im2, (x, y))
	if tesselation == "diagonal":
		x, y = round(maxsize / 2), round(maxsize / 2)
		grid_img = grid_img.resize((round(maxsize * sqrt(2)), round(maxsize * sqrt(2))), resample=Image.Resampling.LANCZOS).rotate(45, resample=Image.Resampling.BICUBIC, expand=True).crop((x, y, x + maxsize, y + maxsize))
	return grid_img

qr_map = {
	2: dict(
		size=32,
		n=42,
		n2=46,
		version=3,
	),
	4: dict(
		size=64,
		n=124,
		n2=124,
		version=11,
	),
	8: dict(
		size=128,
		n=596,
		n2=0,
		version=27,
	),
}

def encode_fountain(data, qr_scale=2, maxsize=512, channels=1, redundancy=6):
	buffer = 2
	try:
		qr_info = qr_map[qr_scale]
	except KeyError:
		raise NotImplementedError(qr_scale)
	qr_size = qr_info["size"]
	n = qr_info["n"]
	mcount = (maxsize // buffer // qr_size) ** 2 * channels
	# print("Maximum chunks:", mcount)
	count = len(data) // n + 1
	# print("Required chunks:", count)
	if len(data) > 65535 or count > mcount:
		raise OverflowError("Input exceeds maximum representable size.")
	used = (maxsize // (qr_size * 2) * 2) ** 2 * channels
	# print("Used chunks:", used)
	def iter_fountain():
		from raptorq import Encoder
		encoder = Encoder.with_defaults(data, n)
		packets = encoder.get_encoded_packets(used + 1)
		for i, packet in enumerate(packets):
			head = len(data).to_bytes(2, "little") + qr_scale.to_bytes(2, "little")
			header = rsc.encode(head)
			text = header + packet
			yield text
	return iter_fountain()

def decode_fountain(fountain):
	decoder = None
	target_size, target_n = None, None
	from raptorq import Decoder
	seen = set()
	i = 0
	for i, data in enumerate(decode_qr(fountain, channels=len(CHANNELS)), 1):
		if not data:
			continue
		header, packet = data[:5], data[5:]
		head = rsc.decode(header)[0]
		size, qr_scale = int.from_bytes(head[:2], "little"), int.from_bytes(head[2:], "little")
		qr_info = qr_map[qr_scale]
		n = qr_info["n"]
		n2 = qr_info["n2"]
		if not decoder:
			target_size = size
			target_n = n
			decoder = Decoder.with_defaults(size, n)
		elif size != target_size or n != target_n:
			raise ValueError("Data size mismatch!")
		if packet in seen:
			continue
		seen.add(packet)
		for j in range(0, len(packet), n2):
			p = packet[j:j + n2]
			assert len(p) == n2, (len(p), n2)
			result = decoder.decode(p)
			if result:
				logging.info(f"Decode successful after {len(seen)}/{i} readable chunks")
				return decode_message(result)
	if not i:
		raise FileNotFoundError("Decode unsuccessful: No valid chunks found")
	raise FileNotFoundError(f"Decode unsuccessful after {len(seen)}/{i} readable chunks")

def forceload(image):
	if isinstance(image, np.ndarray):
		image = Image.fromarray(image, mode="RGB" if image.shape[-1] == 3 else "L")
	elif isinstance(image, (str, io.IOBase)):
		image = Image.open(image)
	elif isinstance(image, (bytes, memoryview)):
		image = Image.open(io.BytesIO(image))
	elif not isinstance(image, Image.Image):
		raise TypeError("`image` should be an instance of `PIL.Image.Image`, `np.ndarray`, `bytes`, `memoryview`, `io.IOBase` or `os.PathLike`.")
	return image

def encode_image(image, message, redundancy=6, dither_wrap=4, strength=1, compress=True, debug=None):
	image = forceload(image)
	if debug and not os.path.exists(debug):
		os.mkdir(debug)
	if debug:
		for fn in os.listdir(debug):
			os.remove(debug + "/" + fn)
		image.save(f"{debug}/initial.png")
	size = image.size
	maxsize = min(2048, min(image.size))
	A = None
	if image.mode != "RGB":
		if "A" in image.mode:
			A = image.getchannel("A")
		image = image.convert("RGB")
	# image = batch_barcodes(fountain, image, strength=strength)
	rgb = np.array(image, dtype=np.float32)
	rgb *= 1 / 255
	if compress:
		data = encode_message(message)
	else:
		data = message
	# print("Encoded message:", data)

	def encode_part(imt, depth=256, start=0):
		for i, arr in enumerate(imt, start):
			if i not in CHANNELS:
				continue
			fountain = encode_fountain(data, qr_scale=2, maxsize=maxsize, channels=1, redundancy=redundancy)
			mult = floor(sqrt(size[0] * size[1] / 1048576))
			if i == 1:
				if size[0] >= 2048:
					tesselation = "pythagorean"
					tile = 1 * mult
				else:
					tesselation = "plain"
					tile = 4 * mult
			elif i == 2:
				tesselation = "diagonal"
				tile = 4 * mult
			else:
				tesselation = "pythagorean"
				tile = 2 * mult
			steg = batch_images(fountain, maxsize, tesselation, tile, seed=i)
			if steg.size != size:
				steg = steg.resize(size, resample=Image.Resampling.NEAREST)
			stegmap = np.asanyarray(steg, dtype=np.uint8).T
			if debug:
				im2 = Image.fromarray(stegmap * 255, "L")
				im2.save(f"{debug}/mask{i}.png")
			if i & 1:
				omap = np.logical_not(stegmap)
			else:
				omap = stegmap.view(bool)
			arr *= depth
			arr[omap] -= 0.5
			arr = np.round(arr, out=arr)
			arr[omap] += 0.5
			arr *= 1 / depth
			np.clip(arr, 0, 1, out=arr)
			if debug:
				a = arr.T * 255
				# print(a.shape)
				a = quantise_into(a, clip=(0, 255))
				im = Image.fromarray(a, mode="L")
				im.save(f"{debug}/{i}.png")
	import cv2

	rand = np.random.random_sample(rgb.shape)
	rand -= 0.5
	rand *= strength / 72
	rgb += rand

	yuv = cv2.cvtColor(rgb, cv2.COLOR_RGB2YCrCb, dst=rgb)
	encode_part(yuv.T, depth=40 / strength, start=0)
	rgb = cv2.cvtColor(yuv, cv2.COLOR_YCrCb2RGB, dst=rgb)

	np.clip(rgb, 0, 1, out=rgb)

	hls = cv2.cvtColor(rgb, cv2.COLOR_RGB2HLS, dst=rgb)
	encode_part(hls.T, depth=64 / strength, start=3)
	rgb = cv2.cvtColor(hls, cv2.COLOR_HLS2RGB, dst=rgb)

	rgb *= 255
	rgb = quantise_into(rgb, clip=(0, 255), checker=dither_wrap)
	im = Image.fromarray(rgb, mode="RGB")
	if A and np.min(A) < 254:
		im.putalpha(A)

	if debug:
		im.save(f"{debug}/fountain.png")
	return im

def decode_image(image, strength=1, debug=None):
	image = forceload(image)
	if debug and not os.path.exists(debug):
		os.mkdir(debug)
	if debug:
		for fn in os.listdir(debug):
			os.remove(debug + "/" + fn)
		image.save(f"{debug}/fountain.png")
	import cv2

	def chunking(im):
		width, height = im.size
		rgb = np.array(im, dtype=np.float32)
		rgb *= 1 / 255
		# rgb = cv2.GaussianBlur(rgb, (3, 3), 0, dst=rgb)
		# rgb = cv2.resize(rgb, (width // 2, height // 2), interpolation=cv2.INTER_AREA)
		ims = []
		def decode_part(imt, depth=256, start=0):
			for i, arr in enumerate(imt, start):
				if i not in CHANNELS:
					continue
				if debug:
					im2 = Image.fromarray((arr.T * 255).astype(np.uint8), "L")
					im2.save(f"{debug}/map-{i}.png")

				arr *= depth
				arr %= 1
				arr -= 0.5
				np.abs(arr, out=arr)
				if not i & 1:
					np.subtract(0.5, arr, out=arr)
				a = arr.T * 2 * 255
				# a = cv2.GaussianBlur(a, (3, 3), 0, dst=a)
				square = int(np.sqrt(width * height))
				if i == 4:
					small = min(2048, max(square // 2, width, height))
				else:
					small = min(1024, max(square // 4, width // 2, height // 2))
				a = cv2.resize(a, (small, small), interpolation=cv2.INTER_AREA)
				im = Image.fromarray(a.astype(np.uint8), mode="L")
				if debug:
					im.save(f"{debug}/{i}.png")
				ims.append(im)
				yield im

		yuv = cv2.cvtColor(rgb, cv2.COLOR_RGB2YCrCb)
		yield from decode_part(yuv.T, depth=40 / strength, start=0)

		hls = cv2.cvtColor(rgb, cv2.COLOR_RGB2HLS)
		yield from decode_part(hls.T, depth=64 / strength, start=3)

	return decode_fountain(chunking(image))


import argparse
parser = argparse.ArgumentParser()
parser.add_argument("image_path")
parser.add_argument("message", default="")
parser.add_argument("-s", "--strength", default=1, required=False)
parser.add_argument("-c", "--compress", action="store_true", default=True, required=False)
parser.add_argument("-e", "--encode", nargs="?", const=True, default=None, required=False)
args = parser.parse_args()

if args.encode is not None:
    from .ectoplasm import encode_image
    im = encode_image(args.image_path, args.message, strength=args.strength, compress=args.compress)
    if not isinstance(args.encode, str):
        args.encode = (p := args.image_path.rsplit(".", 1))[0] + "~." + p[1]
    im.save(args.encode)
    print(f'Output successfully written to "{args.encode}".')
else:
    from .ectoplasm import decode_image
    message = decode_image(args.image_path, strength=args.strength)
    if args.message:
        assert message == args.message.encode("utf-8"), "Decoded message does not match."
    else:
        import sys
        sys.stdout.buffer.write(message)


import io
import logging
import os
import random
import orjson
import numpy as np
from PIL import Image, ImageChops, ImageEnhance, ImageOps
import ectoplasm

strength = 12

def predicate(func, fountain, message, debug=None):
	decoded = ectoplasm.decode_image(func(fountain), strength=strength, debug=debug)
	if decoded != message:
		logging.warning((decoded + b" != " + message).decode("utf-8", "replace"))
	return decoded == message

# Directly decode (as control)
def layer_direct(fountain):
	logging.info("Direct test:")
	return fountain
# Flip randomly (left-right, up-down, transpose etc)
def layer_flip(fountain):
	logging.info("Flip test:")
	return fountain.transpose(random.randint(0, 6))
# Invert all colours in the image
def layer_invert(fountain):
	logging.info("Invert test:")
	return ImageOps.invert(fountain)
# Scroll image randomly left/right/up/down, up to 50% of the image's original size
def layer_translate(fountain):
	logging.info("Translation test:")
	return ImageChops.offset(fountain, round((random.random() - 0.5) * fountain.width), round((random.random() - 0.5) * fountain.height))
# Blank a random section of the image, up to 50% of the image's original size
def layer_blank(fountain):
	logging.info("Blank test:")
	f = fountain.convert("RGBA")
	a = random.random() * 0.25 + 0.5
	b = random.random() * 0.25 + 0.5
	x, y = round(random.random() * (1 - a) * fountain.width), round(random.random() * (1 - b) * fountain.height)
	w, h = round(a * fountain.width), round(b * fountain.height)
	temp = Image.new("RGBA", (w, h), color=(0, 0, 0, 0))
	f.paste(temp, (x, y))
	return f
# Rotate image at any random angle
def layer_rotate(fountain):
	logging.info("Rotation test:")
	return fountain.rotate(random.random() * 360, expand=False, resample=Image.Resampling.BICUBIC)
# Skew image at any random ratio in both axes, up to 50% of the image's original size
def layer_skew(fountain):
	logging.info("Skew test:")
	x = random.random() - 0.5
	y = random.random() - 0.5
	tx = -x * fountain.width / 2
	ty = -y * fountain.height / 2
	return fountain.transform(fountain.size, Image.AFFINE, (1, x, tx, y, 1, ty), resample=Image.Resampling.BICUBIC)
# Crop image to minimum 25% of the image's original size
def layer_crop(fountain):
	logging.info("Crop test:")
	a = random.random() * 0.25 + 0.5
	b = random.random() * 0.25 + 0.5
	x, y = round(random.random() * (1 - a) * fountain.width), round(random.random() * (1 - b) * fountain.height)
	w, h = round(a * fountain.width), round(b * fountain.height)
	return fountain.crop((x, y, x + w, y + h))
# Hueshift image by a random factor
def layer_hueshift(fountain):
	logging.info("Hueshift test:")
	spl = fountain.convert("HSV").split()
	return Image.merge("HSV", (spl[0].point(lambda x: x + random.randint(1, 255) & 255), spl[1], spl[2])).convert("RGB")
# Resize image by a random factor, from 25% to 225% of the image's original size
def layer_resize(fountain):
	logging.info("Resize test:")
	return fountain.resize((round((random.random() + 0.5) * fountain.width), round((random.random() + 0.5) * fountain.height)), resample=Image.Resampling.BICUBIC)
# Performs the magik filter on the image, up to 33% of the image's original size
def quad_as_rect(quad):
	if quad[0] != quad[2]:
		return False
	if quad[1] != quad[7]:
		return False
	if quad[4] != quad[6]:
		return False
	if quad[3] != quad[5]:
		return False
	return True
def quad_to_rect(quad):
	assert(len(quad) == 8)
	assert(quad_as_rect(quad))
	return (quad[0], quad[1], quad[4], quad[3])
def rect_to_quad(rect):
	assert(len(rect) == 4)
	return (rect[0], rect[1], rect[0], rect[3], rect[2], rect[3], rect[2], rect[1])
def shape_to_rect(shape):
	assert(len(shape) == 2)
	return (0, 0, shape[0], shape[1])
def griddify(rect, w_div, h_div):
	w = rect[2] - rect[0]
	h = rect[3] - rect[1]
	x_step = w / float(w_div)
	y_step = h / float(h_div)
	y = rect[1]
	grid_vertex_matrix = []
	for _ in range(h_div + 1):
		grid_vertex_matrix.append([])
		x = rect[0]
		for _ in range(w_div + 1):
			grid_vertex_matrix[-1].append([int(x), int(y)])
			x += x_step
		y += y_step
	grid = np.array(grid_vertex_matrix)
	return grid
def distort_grid(org_grid, max_shift):
	new_grid = np.copy(org_grid)
	x_min = np.min(new_grid[:, :, 0])
	y_min = np.min(new_grid[:, :, 1])
	x_max = np.max(new_grid[:, :, 0])
	y_max = np.max(new_grid[:, :, 1])
	new_grid += np.random.randint(-max_shift, max_shift + 1, new_grid.shape)
	new_grid[:, :, 0] = np.maximum(x_min, new_grid[:, :, 0])
	new_grid[:, :, 1] = np.maximum(y_min, new_grid[:, :, 1])
	new_grid[:, :, 0] = np.minimum(x_max, new_grid[:, :, 0])
	new_grid[:, :, 1] = np.minimum(y_max, new_grid[:, :, 1])
	return new_grid
def grid_to_mesh(src_grid, dst_grid):
	assert(src_grid.shape == dst_grid.shape)
	mesh = []
	for i in range(src_grid.shape[0] - 1):
		for j in range(src_grid.shape[1] - 1):
			src_quad = [
				src_grid[i, j, 0], src_grid[i, j, 1],
				src_grid[i + 1, j, 0], src_grid[i + 1, j, 1],
				src_grid[i + 1, j + 1, 0], src_grid[i + 1, j + 1, 1],
				src_grid[i, j + 1, 0], src_grid[i, j + 1, 1],
			]
			dst_quad = [
				dst_grid[i, j, 0], dst_grid[i, j, 1],
				dst_grid[i + 1, j, 0], dst_grid[i + 1, j, 1],
				dst_grid[i + 1, j + 1, 0], dst_grid[i + 1, j + 1, 1],
				dst_grid[i, j + 1, 0], dst_grid[i, j + 1, 1],
			]
			dst_rect = quad_to_rect(dst_quad)
			mesh.append([dst_rect, src_quad])
	return list(mesh)
def layer_magik(fountain):
	logging.info("Magik test:")
	dst_grid = griddify(shape_to_rect(fountain.size), 4, 4)
	src_grid = distort_grid(dst_grid, np.sqrt(fountain.width * fountain.height) // 4 // 3)
	mesh = grid_to_mesh(src_grid, dst_grid)
	return fountain.transform(fountain.size, Image.Transform.MESH, mesh, resample=Image.Resampling.NEAREST)
# Convert image to jpeg and back
def layer_jpeg(fountain):
	logging.info("JPEG test:")
	b = io.BytesIO()
	fountain.save(b, format="jpeg", quality=90)#random.randint(80, 90))
	b.seek(0)
	return Image.open(b)
# Brighten or darken image by a random factor, up to 50% above or below
def layer_brighten(fountain):
	logging.info("Brightness test:")
	return ImageEnhance.Brightness(fountain).enhance(random.random() + 0.5)

available = {
	"decode-direct": layer_direct,
	"decode-flip": layer_flip,
	"decode-invert": layer_invert,
	"decode-translate": layer_translate,
	"decode-blank": layer_blank,
	"decode-rotate": layer_rotate,
	"decode-skew": layer_skew,
	"decode-crop": layer_crop,
	"decode-hueshift": layer_hueshift,
	"decode-resize": layer_resize,
	"decode-magik": layer_magik,
	"decode-brighten": layer_brighten,
	"decode-jpeg": layer_jpeg,
}

if not os.path.exists("tests"):
	os.mkdir("tests")

test_text = "tests/texts.txt"
with open(test_text, "rb") as f:
	texts = f.read().splitlines()
test_images = "tests/images"
images = [test_images + "/" + t for t in os.listdir(test_images)]

stats = {k: [0, 0] for k in available}
statfile = "tests/stats.json"
if os.path.exists(statfile):
	with open(statfile, "rb") as f:
		stats.update(orjson.loads(f.read()))

for i in range(1000):
	im = None
	while not im:
		path = random.choice(images)
		im = Image.open(path)
		if im.width * im.height < 1048576:
			im = None
	message = b""
	while not message.strip():
		message = random.choice(texts)
	fountain = ectoplasm.encode_image(im, message, redundancy=1, dither_wrap=2, strength=strength, debug="tests/encode")
	transform = random.choice(tuple(available))
	try:
		assert predicate(available[transform], fountain, message, debug="tests/" + transform)
	except Exception as ex:
		logging.error("FAIL: " + repr(ex) + ": " + path + ", " + transform + "; " + message.decode("utf-8", "replace"))
	else:
		stats[transform][0] += 1
	stats[transform][1] += 1
	with open(statfile, "wb") as f:
		f.write(orjson.dumps(stats))